/** @file rawtos2font.c
 *
 * A converter that creates s2font bitmap font files from raw bitmap
 * fonts.
 *
 * Copyright 2004 Petteri Kangaslampi
 *
 * See license.txt for full copyright and license information.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* Type definitions for unsigned 8, 16, and 32-bit integers. */
typedef unsigned char U8;
typedef unsigned short U16;
typedef unsigned int U32;


const U32 s2fontMagic = 0x53324f4e; /* S2FN */
const U32 formatVersion = 0x00010000;
const int numChars = 256;


/* Usage message */
const char *usage =
"Usage: rawtos2font sourcefile destfile width height encoding\n"
"\n"
"        sourcefile   source raw bitmap file\n"
"        destfile     destination s2font file\n"
"        width        character width in pixels\n"
"        height       character height in pixels\n"
"        encoding     encoding, typically CP437 for DOS fonts\n";


/* Exit with error */
void Error(char *msg) {
    fprintf(stderr, "raw2s2font: %s\n", msg);
    exit(EXIT_FAILURE);
}



/* Write a byte */
void WriteU8(FILE *file, U8 val) {
    if ( fwrite(&val, 1, 1, file) != 1 ) {
        Error("Error writing to output file");
    }
}


/* Write a 16-bit unsigned integer in network byte order */
void WriteU16(FILE *file, U16 val) {
    U8 data[2];
    data[0] = (U8) ((val >> 8) & 0xff);
    data[1] = (U8) (val & 0xff);
    if ( fwrite(data, 2, 1, file) != 1 ) {
        Error("Error writing to output file");
    }
}


/* Write a 32-bit unsigned integer in network byte order */
void WriteU32(FILE *file, U32 val) {
    U8 data[4];
    data[0] = (U8) ((val >> 24) & 0xff);
    data[1] = (U8) ((val >> 16) & 0xff);
    data[2] = (U8) ((val >> 8) & 0xff);
    data[3] = (U8) (val & 0xff);
    if ( fwrite(data, 4, 1, file) != 1)  {
        Error("Error writing to output file");
    }
}


int main(int argc, char *argv[]) {

    FILE *sourceFile, *destFile;
    U8 *sourceFont;
    int sourceLen;
    int charWidth, charHeight;
    int byteWidth;
    char *encoding;
    int c, x, y;
    U8 *s;

    if ( argc != 6 ) {
        fputs(usage, stderr);
        return EXIT_FAILURE;
    }

    /* Get parameters and calculate font size */
    charWidth = atoi(argv[3]);
    charHeight = atoi(argv[4]);    
    encoding = argv[5];
    byteWidth = (charWidth+7) / 8;

    /* Open files */
    if ( (sourceFile = fopen(argv[1], "rb")) == NULL ) {
        Error("Unable to open source file");
    }
    if ( (destFile = fopen(argv[2], "wb")) == NULL ) {
        Error("Unable to open destination file");
    }    

    /* Check source length and read file */
    fseek(sourceFile, 0, SEEK_END);
    sourceLen = ftell(sourceFile);
    fseek(sourceFile, 0, SEEK_SET);
    if ( sourceLen != (numChars * charHeight * byteWidth) ) {
        Error("Bad source file");
    }
    sourceFont = (U8*) malloc(sourceLen);
    fread(sourceFont, sourceLen, 1, sourceFile);


    /* Write font file header */
    WriteU32(destFile, s2fontMagic);
    WriteU32(destFile, formatVersion);
    WriteU32(destFile, (U32) 256);
    WriteU32(destFile, (U32) charWidth);
    WriteU32(destFile, (U32) charHeight);
    if ( fwrite(encoding, strlen(encoding)+1, 1, destFile) != 1 ) {
        Error("Error writing to output file");
    }

    /* Write font data */
    s = sourceFont;
    for ( c = 0; c < 256; c++ ) {
        WriteU16(destFile, (U16) (c & 0xffff));
        for ( y = 0; y < charHeight; y++ ) {
            for ( x = 0; x < byteWidth; x++ ) {
                WriteU8(destFile, *s++);
            }
        }
    }

    /* Free resources */
    free(sourceFont);
    fclose(destFile);
    fclose(sourceFile);

    return 0;
}
