/*    ttywindow.cpp
 *
 * TTY window for Nethack for Symbian OS
 *
 * Copyright 2004 Petteri Kangaslampi
 *
 * NetHack may be freely redistributed.  See license for details.
*/

#include "ttywindow.h"
#include <string.h>

_LIT(KAssertPanic, "ttywindow.cpp");
#define assert(x) __ASSERT_ALWAYS(x, User::Panic(KAssertPanic, __LINE__))

_LIT(KPanic, "TtyWindow");
static const TInt KPanicOutOfMemory = 1;

#define DEBUGPRINT
#ifdef DEBUGPRINT
#include <e32svr.h>
#define PRINT(x) RDebug::Print x
#else
#define PRINT(x)
#endif

_LIT(KSpace, " ");


// Factory
CTtyWindow *CTtyWindow::NewL(RUiSession &aSession) {
    CTtyWindow *self = new (ELeave) CTtyWindow(aSession);
    CleanupStack::PushL(self);
    self->ConstructL();
    CleanupStack::Pop(self);
    return self;
}


// Constructor
CTtyWindow::CTtyWindow(RUiSession &aSession)
    : iUiSession(aSession),
      iWidth(80),
      iHeight(24),
      iFgColor(KRgbGray),
      iBgColor(KRgbBlack) {
    // FIXME: Hardwired terminal size
}


// Destructor
CTtyWindow::~CTtyWindow() {
    delete iBuffer;
}


// Second-phase constructor
void CTtyWindow::ConstructL() {
}


// Get window size
void CTtyWindow::GetWindowSize(TInt &aWidth, TInt &aHeight) {
    aWidth = iWidth;
    aHeight = iHeight;
}


// Write text
void CTtyWindow::Write(const char *aText, TInt aNumCharacters) {

    // First convert the text to a descriptor
    if ( aNumCharacters == -1 ) {
        aNumCharacters = strlen(aText);
    }
    HBufC *buf = TextBuffer(aNumCharacters);
    TPtr ptr = buf->Des();
    TPtrC8 ptr8((const TUint8*)aText, aNumCharacters);
    ptr.Copy(ptr8);

    // Build colors from attributes
    TRgb fg, bg;
    ScreenColors(fg, bg);

    // Draw the text on screen, taking care of wraparound and some control
    // characters
    TInt pos = 0;
    TInt start = 0;
    while ( pos < aNumCharacters ) {

        TInt error = KErrNone;

        // Handle control characters
        if ( ptr[pos] < 32 ) {
            // Write current
            if ( pos > start ) {
                DoWrite(ptr.Mid(start, pos-start), fg, bg);
            }

            switch ( ptr[pos] ) {
                case '\b': { // backspace
                    if ( iX > 0 ) {
                        iX--;
                    }
                    error = iUiSession.DrawText(KSpace, iX, iY, fg, bg);
                    if ( error != KErrNone ) {
                        PRINT((_L("DrawText failed with %d"), error));
                        Exit(error);
                    }
                    break;
                }

                case '\n': // CR
                case '\r': { // LF
                    iY++;
                    if ( iY >= iHeight ) {
                        // FIXME: Scroll
                        iY = 0;
                    }
                    iX = 0;
                    break;
                }
            }

            pos++;
            start = pos;
        } else {
            pos++;
        }
    }

    // Draw the rest
    if ( pos > start ) {
        DoWrite(ptr.Mid(start, pos-start), fg, bg);
    }

    // Move the cursor to the final position
    SetCursor(iX, iY);
}


// Draw text at a position
void CTtyWindow::DrawText(TInt aX, TInt aY, const char *aText,
                          TInt aNumCharacters) {
    // First convert the text to a descriptor
    if ( aNumCharacters == -1 ) {
        aNumCharacters = strlen(aText);
    }
    HBufC *buf = TextBuffer(aNumCharacters);
    TPtr ptr = buf->Des();
    TPtrC8 ptr8((const TUint8*)aText, aNumCharacters);
    ptr.Copy(ptr8);

    // Build colors from attributes
    TRgb fg, bg;
    ScreenColors(fg, bg);

    // Draw the text on screen
    TInt error = iUiSession.DrawText(ptr, aX, aY, fg, bg);
    if ( error != KErrNone ) {
        PRINT((_L("DrawText failed with %d"), error));
        Exit(error);
    }
}


// Read keypress
TKeyCode CTtyWindow::GetKey() {
    TKeyCode key;
    TUint32 modifiers = 0;
    GetKey(key, modifiers);
    return key;
}


// Read keypress
void CTtyWindow::GetKey(TKeyCode &aKey, TUint32 aModifiers) {
    TInt error = iUiSession.GetKey(aKey, aModifiers);
    if ( error != KErrNone ) {
        PRINT((_L("GetKey failed with %d"), error));
        Exit(error);
    }    
}


// Set text position
void CTtyWindow::SetPos(TInt aX, TInt aY) {
    iX = aX;
    iY = aY;
    if ( iX < 0 ) {
        iX = 0;
    } else if ( iX >= iWidth ) {
        iX = iWidth - 1;
    }
    if ( iY < 0 ) {
        iY = 0;
    } else if ( iY >= iHeight ) {
        iY = iHeight - 1;
    }
    SetCursor(iX, iY);
}


// Get text position
void CTtyWindow::GetPos(TInt &aX, TInt &aY) {
    aX = iX;
    aY = iY;
}


// Set foreground color
void CTtyWindow::SetFgColor(TRgb aColor) {
    iFgColor = aColor;
}


// Set background color
void CTtyWindow::SetBgColor(TRgb aColor) {
    iBgColor = aColor;
}


// Set inverse video
void CTtyWindow::SetInverse(TBool aInverse) {
    iInverse = aInverse;
}


// Clear to end of line
void CTtyWindow::ClearEndOfLine() {
    // FIXME: Make more efficient?
    TInt len = iWidth - iX;
    if ( len == 0 ) {
        return;
    }

    HBufC *buf = TextBuffer(len);
    TPtr ptr = buf->Des();
    ptr.Fill(' ', len);

    TRgb fg, bg;
    ScreenColors(fg, bg);
    
    TInt error = iUiSession.DrawText(ptr, iX, iY, fg, bg);
    if ( error != KErrNone ) {
        PRINT((_L("DrawText failed with %d"), error));
        Exit(error);
    }    
}


// Clear screen
void CTtyWindow::ClearScreen() {
    // FIXME: Make more efficient!
    HBufC *buf = TextBuffer(iWidth);
    TPtr ptr = buf->Des();
    ptr.Fill(' ', iWidth);

    TRgb fg, bg;
    ScreenColors(fg, bg);

    for ( TInt y = 0; y < iHeight; y++ ) {
        TInt error = iUiSession.DrawText(ptr, 0, y, fg, bg);
        if ( error != KErrNone ) {
            PRINT((_L("DrawText failed with %d"), error));
            Exit(error);
        }
    }
}


// Move the cursor
void CTtyWindow::SetCursor(TInt aX, TInt aY) {
    TInt error = iUiSession.SetCursor(aX, aY);
    if ( error != KErrNone ) {
        PRINT((_L("SetCursor failed with %d"), error));
        Exit(error);
    }
}


// Build final colors
void CTtyWindow::ScreenColors(TRgb &aFgColor, TRgb &aBgColor) {
    aFgColor = iFgColor;
    aBgColor = iBgColor;
    if ( iInverse ) {
        aFgColor ^= TRgb(0xfffff);
        aBgColor ^= TRgb(0xfffff);
    }
}
   

// Build a text buffer
HBufC *CTtyWindow::TextBuffer(TInt aLength) {
    if ( iBuffer ) {
        if ( iBuffer->Length() >= aLength ) {
            return iBuffer;
        }
        delete iBuffer;
        iBuffer = NULL;
    }
    
    iBuffer = HBufC::New(aLength);
    if ( !iBuffer ) {
        User::Panic(KPanic, KPanicOutOfMemory);
    }

    return iBuffer;
}


// Exit the nethack process immediately
void CTtyWindow::Exit(TInt aReason) {
    PRINT((_L("CTtyWindow::Exit(%d"), aReason));
    User::Exit(aReason);
}


// Draws text
void CTtyWindow::DoWrite(const TDesC &aText, TRgb aFgColor, TRgb aBgColor) {

    TInt length = aText.Length();
    TInt pos = 0;
    
    while ( pos < length ) {

        TInt printNow = length - pos;
        if ( printNow > (iWidth - iX) ) {
            printNow = iWidth - iX;
        }
        TInt error = iUiSession.DrawText(aText.Mid(pos, printNow), iX, iY,
                                         aFgColor, aBgColor);
        if ( error != KErrNone ) {
            PRINT((_L("DrawText failed with %d"), error));
            Exit(error);
        }

        pos += printNow;
        iX += printNow;
        if ( iX >= iWidth ) {
            iX = 0;
            iY++;
            if ( iY >= iHeight ) {
                // FIXME: Scroll!
                iY = 0;
            }
        }
    }    
}
