/*    ttywindow.h
 *
 * TTY window for Nethack for Symbian OS
 *
 * Copyright 2004 Petteri Kangaslampi
 *
 * NetHack may be freely redistributed.  See license for details.
*/

#ifndef __TTYWINDOW_H__
#define __TTYWINDOW_H__

#include "uisession.h"


/**
 * TTY window for Nethack. The TTY window maintains current cursor position and
 * attribute status, and provides a simple interface to the TTY implementation.
 *
 * If the UI server is terminated, CTtyWindow will terminate Nethack
 * immediately.
 */
class CTtyWindow : public CBase {
public:
    /** 
     * Factory method.
     * 
     * @param aSession The UI server session to use
     * 
     * @return A new CTtyWindow instance
     */
    static CTtyWindow *NewL(RUiSession &aSession);

    /** 
     * Destructor.
     */
    ~CTtyWindow();

    /** 
     * Gets the TTY window size
     * 
     * @param aWidth Window width in characters
     * @param aHeight Window height in characters
     */
    void GetWindowSize(TInt &aWidth, TInt &aHeight);

    /** 
     * Writes text on the UI window in the current text position. The
     * text position and the cursor are moved to the end of the text.
     * 
     * @param aText The text to write
     * @param aNumCharacters The number of characters to write. By default,
     *                       the entire text is written.
     */
    void Write(const char *aText, TInt aNumCharacters=-1);

    /** 
     * Draws text on the UI window at a specified position. The current text
     * position is not modified.
     * 
     * @param aX Text X-coordinate
     * @param aY Text Y-coordinate
     * @param aText Text
     * @param aNumCharacters The number of characters to write. By default,
     *                       the entire text is written.
     */
    void DrawText(TInt aX, TInt aY, const char *aText, TInt aNumCharacters=-1);

    /** 
     * Reads a keypress, blocking until a keypress is available.
     * 
     * @return Key code
     */
    TKeyCode GetKey();

    /** 
     * Reads a keypress, blocking until a keypress is available
     * 
     * @param aKey Key code
     * @param aModifiers Modifiers for the key
     */
    void GetKey(TKeyCode &aKey, TUint32 aModifiers);

    /** 
     * Sets the text position. Also moves the cursor.
     * 
     * @param aX Position X-coordinate
     * @param aY Position Y-coordinate
     */
    void SetPos(TInt aX, TInt aY);

    /** 
     * Gets the current text position.
     * 
     * @param aX Position X-coordinate
     * @param aY Position Y-coordinate
     */
    void GetPos(TInt &aX, TInt &aY);

    /** 
     * Sets text foreground color.
     * 
     * @param aColor Text foreground color
     */
    void SetFgColor(TRgb aColor);

    /** 
     * Sets text background color.
     * 
     * @param aColor Text background color
     */
    void SetBgColor(TRgb aColor);

    /** 
     * Enables or disables inverse video.
     * 
     * @param aInverse ETrue for inverse video, EFalse for normal
     */
    void SetInverse(TBool aInverse);

    /** 
     * Clears the current line from the current text position to the end of the
     * line. The text position is not modified.
     */
    void ClearEndOfLine();

    /** 
     * Clears the whole screen. The text position is not modified.
     */
    void ClearScreen();

    /** 
     * Sets the cursor position. To hide the cursor, move it to
     * negative coordinates.
     * 
     * @param aX Cursor X-coordinate
     * @param aY Cursor Y-coordinate
     */
    void SetCursor(TInt aX, TInt aY);

private:
    /** 
     * Constructor.
     * 
     * @param aSession The UI server session to use
     */
    CTtyWindow(RUiSession &aSession);

    /** 
     * Second-phase constructor.
     */
    void ConstructL();

    /** 
     * Exits the nethack process immediately.
     * 
     * @param aReason Exit reason code
     */
    void Exit(TInt aReason);

    /** 
     * Builds final colors from the current colors and attributes.
     * 
     * @param aFgColor Foreground color
     * @param aBgColor Background color
     */
    void ScreenColors(TRgb &aFgColor, TRgb &aBgColor);

    /** 
     * Returns a text buffer of suitable length. The buffer may be
     * reallocated or recycled. The buffer is deallocated automatically when
     * the class is destroyed, and the user must not delete it.
     * 
     * @param aLength Text length in characters.
     * 
     * @return A text buffer
     */
    HBufC *TextBuffer(TInt aLength);

    /** 
     * Actually draws text on the screen, handling wraparound, but not control
     * characters.
     *
     * @param aText Text
     * @param aFgColor Foreground color
     * @param aBgColor Background color
     */
    void DoWrite(const TDesC &aText, TRgb aFgColor, TRgb aBgColor);
    
    
    RUiSession &iUiSession;
    TInt iX, iY; // text position
    TInt iWidth, iHeight; // terminal size
    TRgb iFgColor, iBgColor;
    TBool iInverse;
    HBufC *iBuffer;
};


#endif
