/*    nethackappui.cpp
 *
 * Nethack UI Application UI class
 *
 * Copyright 2002-2004 Petteri Kangaslampi
 *
 * See license.txt for full copyright and license information.
*/

#include <eikenv.h>
#include <eikon.hrh>
#include <ckninfo.h>
#include <cknconf.h>
#include <eikmenup.h>
#include <eikdoc.h>
#include <eikapp.h>
#include <apgwgnam.h>
#include <apgtask.h>
#include <apgcli.h>
#include <apacmdln.h>
#include <e32hal.h>
#include "s2font.h"

#include "nethackappui.h"
#include "nethackappview.h"
#include "nethackui.hrh"
#include <nethack.rsg>

_LIT(KAssertPanic, "nethackappui.cpp");
#define assert(x) __ASSERT_ALWAYS(x, User::Panic(KAssertPanic, __LINE__))

_LIT(KDefaultFont, "vga8x8.s2f");

#ifdef __WINS__
_LIT(KEngineFile, "nethackengine.dll");
#else
_LIT(KEngineFile, "nethackengine.exe");
#endif

const char *const KWelcome1 = "Nethack 3.4.3";
const char *const KWelcome2 = "Built " __DATE__;


//#include <e32svr.h>
//#define DEBUGPRINT(x) RDebug::Print x
#define DEBUGPRINT(x)


CNethackAppUi::CNethackAppUi() {
    iTermWidth = 80;
    iTermHeight = 24;
    iLargeFont = EFalse;
}


void CNethackAppUi::ConstructL() {

    User::CompressAllHeaps();
    BaseConstructL();

    // Determine application installation path
    HBufC *appDll = HBufC::NewLC(KMaxFileName);
    *appDll = iDocument->Application()->AppFullName();
    TParse parsa;
    User::LeaveIfError(parsa.SetNoWild(*appDll, NULL, NULL));
    iAppPath = parsa.DriveAndPath();
    CleanupStack::PopAndDestroy(); // appDll

    // Make sure the path ends with a backslash
    if ( iAppPath[iAppPath.Length()-1] != '\\' ) {
        iAppPath.Append('\\');
    }

    // Load the font to use
    TFileName fontFileName = iAppPath;
    fontFileName.Append(KDefaultFont);
    iFont = CS2Font::NewL(fontFileName);

    iAppView = new (ELeave) CNethackAppView(this, iFont);
    iAppView->ConstructL(ClientRect());
    AddToStackL(iAppView);

    iServer = CUiServer::NewL(*this, *iAppView->Terminal());

    DrawWelcomeMessage();
}


CNethackAppUi::~CNethackAppUi() {
    delete iServer;
    delete iAppView;
    delete iFont;
}


void CNethackAppUi::HandleCommandL(TInt aCommand) {
    
    CEikonEnv *env = CEikonEnv::Static();

    switch (aCommand) {

        case ENethackCmdStartEngine: {
            if ( iGameRunning ) {
                break;
            }
            if ( !iServer ) {
                iServer = CUiServer::NewL(*this, *iAppView->Terminal());    
            }

            iAppView->SetFullScreenL(ETrue);
            
            // Determine the path to the engine EXE
            TFileName engineExe = iAppPath;
            engineExe.Append(KEngineFile);

            // Connect to launch server
            RApaLsSession ls;
            User::LeaveIfError(ls.Connect());
            CleanupClosePushL(ls);    

            // Build a command line to start the engine and start it
            CApaCommandLine *cl = CApaCommandLine::NewLC();
            cl->SetCommandL(EApaCommandRun);
            cl->SetLibraryNameL(engineExe);
            User::LeaveIfError(ls.StartApp(*cl));

            CleanupStack::PopAndDestroy(2);
            iGameRunning = ETrue;
            break;
        }

        case ENethackCmdMemoryStatus:
            ShowMemoryStatusL();
            break;

        case ENethackCmdSafeExit:
            if ( iGameRunning ) {
                CCknInfoDialog::RunDlgLD(
                    *(env->AllocReadResourceLC(R_STR_CLOSE_GAME_TITLE)),
                    *(env->AllocReadResourceLC(R_STR_CLOSE_GAME_INFO)));
                CleanupStack::PopAndDestroy(2); // title, info)
                break;
            }
            iExiting = ETrue;
            Exit();
            break;

	case EEikCmdExit:
            iExiting = ETrue;
            Exit();
            break;
    }    
}


void CNethackAppUi::ShowMemoryStatusL() {
    
    CEikonEnv *env = CEikonEnv::Static();
    HBufC *title = env->AllocReadResourceLC(R_STR_MEMORY_TITLE);
    HBufC *fmt = env->AllocReadResourceLC(R_STR_MEMORY_INFO);
    
    TBuf<128> buf;
    TMemoryInfoV1Buf memory;
    UserHal::MemoryInfo(memory);

    buf.Format(*fmt, memory().iMaxFreeRamInBytes, memory().iFreeRamInBytes);
    CCknInfoDialog::RunDlgLD(*title, buf);

    CleanupStack::PopAndDestroy(2); // fmt, title
}


void CNethackAppUi::DrawWelcomeMessage() {

    iAppView->Terminal()->SetCursor(-1, -1);

    iAppView->Terminal()->ClearTerminal(KRgbBlack);    
    TSize termSize = iAppView->Terminal()->TerminalSize();

    TBuf<64> buf;
    
    buf.Copy(TPtrC8((TUint8*) KWelcome1));
    TInt y = termSize.iHeight/2 - 2;
    TInt x = termSize.iWidth/2 - buf.Length()/2;
    iAppView->Terminal()->DrawText(x, y, buf, EFalse, EFalse,
                                   KRgbWhite, KRgbBlack);

    buf.Copy(TPtrC8((TUint8*) KWelcome2));
    y += 2;
    x = termSize.iWidth/2 - buf.Length()/2;
    iAppView->Terminal()->DrawText(x, y, buf, EFalse, EFalse,
                                   KRgbWhite, KRgbBlack);
}


void CNethackAppUi::MusoSessionStarted() {
    iEikonEnv->InfoMsg(_L("Session started"));
}


void CNethackAppUi::MusoSessionTerminated() {
    if ( !iExiting ) {
        iAppView->SetFullScreenL(EFalse);
        DrawWelcomeMessage();
        iGameRunning = EFalse;
    }
}


void CNethackAppUi::MusoErrorMessage(const TDesC &aMessage) {
    HBufC *title = CEikonEnv::Static()->AllocReadResourceLC(R_STR_ERROR_TITLE);
    CCknInfoDialog::RunDlgLD(*title, aMessage);
    CleanupStack::PopAndDestroy();
}


void CNethackAppUi::MusoFatalError(TInt aError) {
    User::Panic(_L("Fatal error"), aError);
}
