/*    nethackappview.cpp
 *
 * Nethack UI View class
 *
 * Copyright 2002-2004 Petteri Kangaslampi
 *
 * See license.txt for full copyright and license information.
*/

#include <e32std.h>
#include <gdi.h>
#include <eikenv.h>
#include <apgtask.h>
#include <eikspane.h>
#include <eikbtgpc.h>
#include "nethackappui.h"
#include "nethackappview.h"
#include "nethackui.hrh"

_LIT(KAssertPanic, "nethackappview.cpp");
#define assert(x) __ASSERT_ALWAYS(x, User::Panic(KAssertPanic, __LINE__))

static const TInt KTargetTermWidth = 80;
static const TInt KTargetTermHeight = 24;


CNethackAppView::CNethackAppView(CNethackAppUi *aAppUi, CS2Font *aFont) {
    iAppUi = aAppUi;
    iFont = aFont;
}

CNethackAppView::~CNethackAppView() {
    delete iTerminal;
}


void CNethackAppView::ConstructL(const TRect &aRect) {
    CreateWindowL();
    SetRect(aRect);

    // Build the terminal
    iTermRect = Rect();
    iTerminal = CTerminalControl::NewL(*iFont, iTermRect, Window());

    // Adjust terminal size to max 80x24
    ResizeTerminal();
    
    ActivateL();
}


void CNethackAppView::Draw(const TRect & /*aRect*/) const {

    CWindowGc &gc = SystemGc();
    gc.Reset();
    gc.SetClippingRect(Rect());

    // Clear everything outside the terminal
    TRegionFix<5> clearReg(Rect());
    clearReg.SubRect(iTermRect);
    assert(!clearReg.CheckError());
    const TRect *rects = clearReg.RectangleList();
    TInt numRects = clearReg.Count();
    
    gc.SetBrushStyle(CGraphicsContext::ESolidBrush);
    gc.SetBrushColor(KRgbBlack);
    gc.SetPenStyle(CGraphicsContext::ENullPen);
    while ( numRects-- ) {
        gc.DrawRect(*(rects++));
    }
}


TInt CNethackAppView::CountComponentControls() const {
    return 1;
}


CCoeControl *CNethackAppView::ComponentControl(TInt aIndex) const {
    
    switch ( aIndex ) {
        case 0:
            return iTerminal;

        default:
            assert(EFalse);
    }

    return NULL;
}


TKeyResponse CNethackAppView::OfferKeyEventL(const TKeyEvent &aKeyEvent,
                                          TEventCode aType) {

    // Normally we'll just let the terminal control handle the event, but
    // there are some exceptions:
    // - If the user presses Chr without other keys,
    //   we'll launch the character selection dialog.
    if ( (aType == EEventKeyDown) && (aKeyEvent.iCode == EKeyNull) &&
         ((aKeyEvent.iModifiers & EModifierLeftFunc) != 0) ) {
        iChrDown = ETrue;
    } else if ( iChrDown &&
                (aType == EEventKeyUp) &&
                (aKeyEvent.iCode == EKeyNull) &&
                ((aKeyEvent.iModifiers & EModifierLeftFunc) == 0) ) {
        // Chr up/down without other keys
        iChrDown = EFalse;
//        iAppUi->HandleCommandL(ECmdSendSpecialCharacter);
        return EKeyWasConsumed;
    } else {
        iChrDown = EFalse;
    }

    return iTerminal->OfferKeyEventL(aKeyEvent, aType);
}


CTerminalControl *CNethackAppView::Terminal() {
    return iTerminal;
}


void CNethackAppView::ResizeTerminal() {

    if ( !iTerminal ) {
        return;
    }

    // Make terminal as big as possible not exceeding the target size
    TSize termSize = iTerminal->PixelSize(TSize(KTargetTermWidth,
                                                KTargetTermHeight));
    if ( Rect().Width() < termSize.iWidth ) {
        termSize.iWidth = Rect().Width();
    }
    if ( Rect().Height() < termSize.iHeight ) {
        termSize.iHeight = Rect().Height();
    }
    termSize = iTerminal->PixelSize(iTerminal->TerminalSize(termSize));

    // Center terminal on screen
    TInt termX = Rect().iTl.iX + (Rect().Width() - termSize.iWidth) / 2;
    TInt termY = Rect().iTl.iY + (Rect().Height() - termSize.iHeight) / 2;
    iTermRect = TRect(TPoint(termX, termY), termSize);
    
    iTerminal->SetRect(iTermRect);
    DrawDeferred();
}


void CNethackAppView::SetFullScreenL(TBool aFullScreen) {

    CEikonEnv *eikonEnv = CEikonEnv::Static();

    iFullScreen = aFullScreen;
    
    if ( iFullScreen ) {
        // Hide indicator bar and CBA:
        eikonEnv->AppUiFactory()->StatusPane()->MakeVisible(EFalse);
        eikonEnv->AppUiFactory()->ToolBar()->MakeVisible(EFalse);

        // Set view to cover full screen
        SetRect(iAppUi->ClientRect());
        
    } else {
        // Show indicator bar and CBA:
        eikonEnv->AppUiFactory()->StatusPane()->MakeVisible(ETrue);
        eikonEnv->AppUiFactory()->ToolBar()->MakeVisible(ETrue);

        // Set view to cover the remaining area
        SetRect(iAppUi->ClientRect());
    }

    DrawDeferred();
}


void CNethackAppView::SizeChanged() {
    ResizeTerminal();
}
