/*    terminalcontrol.cpp
 *
 * A terminal UI control class
 *
 * Copyright 2002-2004 Petteri Kangaslampi
 *
 * See license.txt for full copyright and license information.
*/

#include <e32std.h>
#include <gdi.h>
#include <eikenv.h>
#include <e32svr.h>
#include <bitstd.h>
#include <bitdev.h>
#include "terminalcontrol.h"
#include "s2font.h"
#include "oneshottimer.h"

_LIT(KPanic, "TerminalControl");

_LIT(KTerminalFont, "Terminal");
_LIT(KTerminalControl, "terminalcontrol.cpp");

#define assert(x) __ASSERT_ALWAYS(x, User::Panic(KTerminalControl, __LINE__))

const TInt KUpdateTimerDelay = 10000;


CTerminalControl *CTerminalControl::NewL(CS2Font &aFont, const TRect &aRect,
                                         RWindow &aContainerWindow) {
    CTerminalControl *self = new (ELeave) CTerminalControl(aFont);
    CleanupStack::PushL(self);
    self->ConstructL(aRect, aContainerWindow);
    CleanupStack::Pop();
    return self;
}



CTerminalControl::CTerminalControl(CS2Font &aFont) :
    iFont(aFont),
    iCursorX(-1),
    iCursorY(-1) {

    TSize size = aFont.FontSize();    
    iFontHeight = size.iHeight;
    iFontWidth = size.iWidth;
}


CTerminalControl::~CTerminalControl() {

    delete iTimer;
    delete iRowBitmapGc;
    delete iRowBitmapDevice;
    delete iRowBitmap;
    delete [] iChars;
    delete [] iAttributes;
    delete [] iDirtyLeft;
    delete [] iDirtyRight;
}


void CTerminalControl::ConstructL(const TRect &aRect,
                                  RWindow &aContainerWindow) {

    SetContainerWindowL(aContainerWindow);
    
    iTimer = COneShotTimer::NewL(TCallBack(UpdateCallBack, this));
    SetRect(aRect);
    Resize();
    AllocateBuffersL();
    Clear(KRgbWhite);
}


void CTerminalControl::SetObserver(MTerminalObserver *aObserver) {
    iObserver = aObserver;
}


void CTerminalControl::RemoveObserver() {
    iObserver = NULL;
}



void CTerminalControl::SetGrayed(TBool aGrayed) {
    iGrayed = aGrayed;
    DrawDeferred();
}


// Set cursor position
void CTerminalControl::SetCursor(TInt aX, TInt aY) {

    // Mark the old cursor position dirty
    if ( (iCursorX >= 0) && (iCursorX < iCharWidth) &&
         (iCursorY >= 0) && (iCursorY < iCharHeight) ) {
        if ( iDirtyLeft[iCursorY] > iCursorX ) {
            iDirtyLeft[iCursorY] = iCursorX;
        }
        if ( iDirtyRight[iCursorY] < iCursorX ) {
            iDirtyRight[iCursorY] = iCursorX;
        }
    }

    // Move the cursor and mark the new position dirty too
    iCursorX = aX;
    iCursorY = aY;
    if ( (iCursorX >= 0) && (iCursorX < iCharWidth) &&
         (iCursorY >= 0) && (iCursorY < iCharHeight) ) {
        if ( iDirtyLeft[iCursorY] > iCursorX ) {
            iDirtyLeft[iCursorY] = iCursorX;
        }
        if ( iDirtyRight[iCursorY] < iCursorX ) {
            iDirtyRight[iCursorY] = iCursorX;
        }
    }

    StartUpdateTimer();
}


// Resizes the terminal in response to size or font change
void CTerminalControl::Resize() {

    TRect rect = Rect();
    TInt width = rect.iBr.iX - rect.iTl.iX;
    TInt height = rect.iBr.iY - rect.iTl.iY;

    assert((iFontWidth > 0) && (iFontHeight > 0));
    assert((width >= iFontWidth) && (height >= iFontHeight));
    iCharWidth = width / iFontWidth;
    iCharHeight = height / iFontHeight;
    if ( iObserver )
        iObserver->TerminalSizeChanged(iCharWidth, iCharHeight);
    TRAPD(error, AllocateBuffersL());
    if ( error != KErrNone ) {
        User::Panic(KPanic, error);
    }
}


// (re)allocate character and attribute buffers
void CTerminalControl::AllocateBuffersL() {
    delete [] iChars;
    iChars = NULL;
    delete [] iAttributes;
    iAttributes = NULL;
    delete iRowBitmapGc;
    iRowBitmapGc = NULL;
    delete iRowBitmapDevice;
    iRowBitmapDevice = NULL;
    delete iRowBitmap;
    iRowBitmap = NULL;
    delete [] iDirtyLeft;
    iDirtyLeft = NULL;
    delete [] iDirtyRight;
    iDirtyRight = NULL;

    iChars = new (ELeave) TText[iCharWidth*iCharHeight];
    iAttributes = new (ELeave) TTerminalAttribute[iCharWidth*iCharHeight];

    iRowBitmap = new CFbsBitmap();
    TDisplayMode displayMode = SystemGc().Device()->DisplayMode();
    User::LeaveIfError(iRowBitmap->Create(TSize(iCharWidth*iFontWidth,
                                                iFontHeight),
                                          displayMode));
    iRowBitmapDevice = CFbsBitmapDevice::NewL(iRowBitmap);
    iRowBitmapGc = CFbsBitGc::NewL();
    iDirtyLeft = new (ELeave) TInt[iCharHeight];
    iDirtyRight = new (ELeave) TInt[iCharHeight];
    Clear(KRgbWhite);
}


// Clear the buffers
void CTerminalControl::Clear(TRgb aBackground) {
    TText *c = iChars;
    TTerminalAttribute *a = iAttributes;
    TInt num = iCharWidth * iCharHeight;
    while ( num-- ) {
        *c++ = ' ';
        a->iFgColor = KRgbBlack;
        a->iBgColor = aBackground;
        a->iBold = EFalse;
        a->iUnderline = EFalse;
        a++;
    }

    num = iCharHeight;
    TInt *l = iDirtyLeft;
    TInt *r = iDirtyRight;
    while ( num-- ) {
        *l++ = iCharWidth;
        *r++ = -1;
    }
}


// Draws text on the terminal window. The coordinates are zero-based
// character coordinates inside the terminal.
void CTerminalControl::DrawText(TInt aX, TInt aY, const TDesC &aText,
                                TBool aBold, TBool aUnderline,
                                TRgb aForeground, TRgb aBackground) {
    // Don't draw text while grayed out
    if ( iGrayed ) {
        return;
    }

    // Check that we are at least partially on screen
    if ( (aX < 0) || (aY < 0) || (aX >= iCharWidth) || (aY >= iCharHeight) ) {
        return;
    }

    // Write the text to our character and attribute buffers
    TInt numChars = aText.Length();
    if ( numChars > (iCharWidth - aX) ) {
        numChars = iCharWidth - aX;
    }
    TText *c = &iChars[aX + aY*iCharWidth];
    TTerminalAttribute *a = &iAttributes[aX + aY * iCharWidth];
    for ( TInt i = 0; i < numChars; i++ ) {
        *c++ = aText[i];
        a->iFgColor = aForeground;
        a->iBgColor = aBackground;
        a->iBold = aBold;
        a->iUnderline = aUnderline;
        a++;
    }

    // Mark the area dirty
    if ( iDirtyLeft[aY] > aX ) {
        iDirtyLeft[aY] = aX;
    }
    if ( iDirtyRight[aY] < (aX + numChars - 1) ) {
        iDirtyRight[aY] = aX + numChars - 1;
    }
    
    StartUpdateTimer();
}


// Clears the terminal with a given background color
void CTerminalControl::ClearTerminal(TRgb aColor) {

    // Clear the buffer
    Clear(aColor);

    // Clear the screen
    CWindowGc &gc = SystemGc();
    gc.Activate(Window());
    gc.SetBrushStyle(CGraphicsContext::ESolidBrush);
    gc.SetBrushColor(aColor);
    gc.SetPenStyle(CGraphicsContext::ENullPen);
    gc.DrawRect(Rect());
    gc.Deactivate();
}


// Returns the current size in characters
TSize CTerminalControl::TerminalSize() {
    return TSize(iCharWidth, iCharHeight);
}


// Calculates a size in characters from a size in pixels
TSize CTerminalControl::TerminalSize(TSize aPixelSize) {
    return TSize((aPixelSize.iWidth + (iFontWidth-1)) / iFontWidth,
                 (aPixelSize.iHeight + (iFontHeight-1)) / iFontHeight);
}


// Calculates a size in pixels from a size in characters
TSize CTerminalControl::PixelSize(TSize aTerminalSize) {
    return TSize(aTerminalSize.iWidth * iFontWidth,
                 aTerminalSize.iHeight * iFontHeight);
}


// Renders the cursor to the current row bitmap. aX is the cursor character
// position, relative to the start of the bitmap
void CTerminalControl::RenderCursor(TInt aX) const {

    TInt x = aX * iFontWidth;
    TInt y = iFontHeight - 2;
    
    assert((x >= 0) && (x < iCharWidth * iFontWidth));

    iRowBitmapGc->Activate(iRowBitmapDevice);

    iRowBitmapGc->Reset();
    iRowBitmapGc->SetDrawMode(CGraphicsContext::EDrawModeXOR);
    iRowBitmapGc->SetPenColor(KRgbWhite);
    iRowBitmapGc->SetBrushStyle(CGraphicsContext::ENullBrush);
    iRowBitmapGc->DrawRect(TRect(x, y, x+iFontWidth, y+2));    
}


// CCoeControl::OfferKeyEventL()
TKeyResponse CTerminalControl::OfferKeyEventL(const TKeyEvent &aKeyEvent,
                                              TEventCode aType) {

    if ( aType == EEventKey ) {
        TUint code = aKeyEvent.iCode;

        // Pass keypress to the observer
        if ( iObserver )
            iObserver->KeyPressed((TKeyCode)code, aKeyEvent.iModifiers);
        return EKeyWasConsumed;
    }

    return EKeyWasNotConsumed;
}


// CCoeControl::Draw()
void CTerminalControl::Draw(const TRect & /*aRect*/) const {

    CWindowGc &gc = SystemGc();
    gc.Reset();
    RWsSession ws = CCoeEnv::Static()->WsSession();

    if ( iGrayed ) {
        gc.SetBrushStyle(CGraphicsContext::ESolidBrush);
        gc.SetBrushColor(TRgb(0xcccccc));
        gc.SetPenStyle(CGraphicsContext::ENullPen);
        gc.DrawRect(Rect());
    }
    else {
        TInt x0 = Rect().iTl.iX;
        TInt y0 = Rect().iTl.iY;
        
        // We'll find the longest possible run of identical attributes on
        // each line and render those at one go. We'll then blit the bitmap
        // for that row.
        for ( TInt y = 0; y < iCharHeight; y++ ) {
            TText *lChars = &iChars[y * iCharWidth];
            TTerminalAttribute *lAttribs = &iAttributes[y * iCharWidth];
            TInt x = 0;

            while ( x < iCharWidth ) {
                TTerminalAttribute a0 = lAttribs[x];
                TTerminalAttribute *a = &lAttribs[x];
                TInt num = 0;
                while ( ((x+num) < iCharWidth) &&
                        (a->iFgColor == a0.iFgColor) &&
                        (a->iBgColor == a0.iBgColor) &&
                        (a->iBold == a0.iBold) &&
                        (a->iUnderline == a0.iUnderline) ) {
                    a++;
                    num++;
                }
                iFont.RenderText(*iRowBitmap, TPtrC(&lChars[x], num),
                                 x * iFontWidth, 0, a0.iFgColor, a0.iBgColor);
                x += num;
            }

            if ( (iCursorY == y) && ((iCursorX >= 0) &&
                                     (iCursorX < iCharWidth)) ) {
                RenderCursor(iCursorX);
            }
            
            gc.BitBlt(TPoint(x0, y0 + y*iFontHeight), iRowBitmap);
            ws.Flush();
        }
    }
}


// Start update timer
void CTerminalControl::StartUpdateTimer() {
    iTimer->After(KUpdateTimerDelay);
}


// Update all dirty areas on the screen
void CTerminalControl::Update() {

    CWindowGc &gc = SystemGc();
    gc.Activate(Window());
    RWsSession ws = CCoeEnv::Static()->WsSession();
    
    TInt x0 = Rect().iTl.iX;
    TInt y0 = Rect().iTl.iY;

    // Go through each line, finding areas that need updating. On those lines
    // we'll first render all runs of identical updates at one go to the
    // bitmap, render the cursor if necessary, and finally blit the bitmap
    // to the screen
    for ( TInt y = 0; y < iCharHeight; y++ ) {

        // Skip lines that are fully up to date
        if ( (iDirtyRight[y] < iDirtyLeft[y]) ) {
            continue;
        }

        // Render text to the bitmap
        TText *lChars = &iChars[y * iCharWidth];
        TTerminalAttribute *lAttribs = &iAttributes[y * iCharWidth];
        TInt left = iDirtyLeft[y];
        TInt right = iDirtyRight[y];
        TInt x = left;
        TInt chars = right - left + 1;
        assert((x >= 0) && (chars <= iCharWidth));

        while ( chars ) {
            TTerminalAttribute a0 = lAttribs[x];
            TTerminalAttribute *a = &lAttribs[x];
            TInt num = 0;
            while ( ((x+num) <= right) &&
                    (a->iFgColor == a0.iFgColor) &&
                    (a->iBgColor == a0.iBgColor) &&
                    (a->iBold == a0.iBold) &&
                    (a->iUnderline == a0.iUnderline) ) {
                a++;
                num++;
            }
            iFont.RenderText(*iRowBitmap, TPtrC(&lChars[x], num),
                             x * iFontWidth, 0, a0.iFgColor, a0.iBgColor);
            x += num;
            chars -= num;
        }

        // Render the cursor if necessary
        if ( (iCursorY == y) && ((iCursorX >= left) &&
                                 (iCursorX <= right)) ) {
            RenderCursor(iCursorX);
        }

        // Blit
        gc.BitBlt(TPoint(x0 + left*iFontWidth, y0 + y*iFontHeight),
                  iRowBitmap,
                  TRect(left*iFontWidth, 0,
                        (right+1)*iFontWidth, iFontHeight));
        ws.Flush();

        // No longer dirty
        iDirtyLeft[y] = iCharWidth;
        iDirtyRight[y] = -1;
    }

    gc.Deactivate();
}


// Update timer callback
TInt CTerminalControl::UpdateCallBack(TAny *aPtr) {
    ((CTerminalControl*)aPtr)->Update();
    return 0;
}


// CCoeControl::SizeChanged()
void CTerminalControl::SizeChanged() {
    Resize();
}


// CCoeControl::InputCapabilities()
TCoeInputCapabilities CTerminalControl::InputCapabilities() const {
    return TCoeInputCapabilities(TCoeInputCapabilities::EAllText |
                                 TCoeInputCapabilities::ENavigation);
}
