/*    terminalcontrol.h
 *
 * A terminal UI control class
 *
 * Copyright 2002,2004 Petteri Kangaslampi
 *
 * See license.txt for full copyright and license information.
*/

#ifndef __TERMINALCONTROL_H__
#define __TERMINALCONTROL_H__

#include <e32std.h>
#include <coecntrl.h>


class MTerminalObserver {

public:
    // Terminal size has been changed
    virtual void TerminalSizeChanged(TInt aWidth, TInt aHeight) = 0;

    // A key has been pressed
    virtual void KeyPressed(TKeyCode aCode, TUint aModifiers) = 0;
};


class CS2Font;
class CFbsBitmap;
class CFbsBitmapDevice;
class CFbsBitGc;
class COneShotTimer;


/**
 * Attributes for a single character on the terminal.
 */
struct TTerminalAttribute {
    TRgb iFgColor;
    TRgb iBgColor;
    TBool iBold;
    TBool iUnderline;
};


/**
 * A terminal UI control class. Manages a terminal control on the screen,
 * taking care of drawing the text and delivering keypresses to the user.
 * The terminal user must implement the MTerminalObserver interface.
 */
class CTerminalControl : public CCoeControl {

public:
    /** 
     * Factory method.
     *
     * @param aFont The font to use
     * @param aRect Initial terminal control rectangle on screen. Use SetRect()
     *              to change the terminal size.
     * @param aContainerWindow The window that contains this control.
     */
    static CTerminalControl *NewL(CS2Font &aFont, const TRect &aRect,
                                  RWindow &aContainerWindow);

    /** 
     * Destructor.
     */
    ~CTerminalControl();

    /** 
     * Sets the terminal observer.
     * 
     * @param aObserver The observer object
     */
    void SetObserver(MTerminalObserver *aObserver);

    /** 
     * Removes the observer.
     */
    void RemoveObserver();

    /** 
     * Draws text on the terminal window. The coordinates are zero-based
     * character coordinates inside the terminal. There is no clipping --
     * the terminal control assumes that the user takes care of this.
     *
     * Cursors can be drawn as text with a different background color.
     * 
     * @param aX Text X-coordinate
     * @param aY Text Y-coordinate
     * @param aText The text to draw
     * @param aBold ETrue if text should be bold (not supported currently)
     * @param aUnderline ETrue if text should be underlined.
     * @param aForeground Text foreground color
     * @param aBackground Text background color. The text is drawn inside a
     *                    solid rectangle colored with this color.
     */
    void DrawText(TInt aX, TInt aY, const TDesC &aText, TBool aBold,
                  TBool aUnderline, TRgb aForeground,
                  TRgb aBackground);

    /** 
     * Clears the terminal by filling it with a background color.
     * 
     * @param aColor The terminal background color
     */
    void ClearTerminal(TRgb aColor);

    /** 
     * Gets the current terminal size in characters.
     * 
     * @return Terminal size in characters
     */
    TSize TerminalSize();

    /** 
     * Calculates the terminal size in characters from a given size in pixels.
     * 
     * @param aPixelSize Size in pixels
     * 
     * @return Size in characters
     */
    TSize TerminalSize(TSize aPixelSize);

    /** 
     * Calculates a correct size in pixels from a given size in characters.
     * 
     * @param aTerminalSize Size in characters
     * 
     * @return Size in pixels
     */
    TSize PixelSize(TSize aTerminalSize);

    /** 
     * Sets the terminal control to be grayed out or normal. The terminal
     * control can be grayed out when there is no connection, and it cannot
     * be redrawn properly.
     * 
     * @param aGrayed True if the terminal should be grayed out.
     */
    void SetGrayed(TBool aGrayed);

    /** 
     * Sets the cursor position. To hide the cursor, move it to
     * negative coordinates.
     *
     * @param aX Cursor X-coordinate
     * @param aY Cursor Y-coordinate
     */
    void SetCursor(TInt aX, TInt aY);


    /** \fn void SetRect(const TRect aRect);
     *
     * Sets the terminal control size and position on screen. The actual size
     * used is rounded down from the argument to the nearest size in full
     * characters.
     *
     * @param aRect New terminal window size and position.
     */

    // CCoeControl methods
    virtual TKeyResponse OfferKeyEventL(const TKeyEvent &aKeyEvent,
                                        TEventCode aType);
    virtual void Draw(const TRect &aRect) const;
    virtual void SizeChanged();
    virtual TCoeInputCapabilities InputCapabilities() const;

private:
    /** 
     * Constructor.
     */
    CTerminalControl(CS2Font &aFont);
    
    /** 
     * Second-phase constructor.
     */
    void ConstructL(const TRect &aRect, RWindow &aContainerWindow);
    
    
    void Resize();
    void Clear(TRgb aBackground);
    void AllocateBuffersL();
    void StartUpdateTimer();
    void DoDraw(CGraphicsContext &aGc, TInt aX, TInt aY, const TDesC &aText,
                TBool aBold, TBool aUnderline, TRgb aForeground,
                TRgb aBackground) const;
    void RenderCursor(TInt aX) const;
    void Update();
    static TInt UpdateCallBack(TAny *aPtr);

    CS2Font &iFont;
    CFbsBitmap *iRowBitmap;
    CFbsBitmapDevice *iRowBitmapDevice;
    CFbsBitGc *iRowBitmapGc;
    TInt iFontWidth, iFontHeight; // font char dimensions in pixels
    MTerminalObserver *iObserver;
    TBool iGrayed;
    TText *iChars;
    TTerminalAttribute *iAttributes;
    TInt iCharWidth, iCharHeight;
    TInt iCursorX, iCursorY;
    COneShotTimer *iTimer;
    TInt *iDirtyLeft;
    TInt *iDirtyRight;
};


#endif
