/*    uisession.cpp
 *
 * UI server session
 *
 * Copyright 2004 Petteri Kangaslampi
 *
 * See license.txt for full copyright and license information.
*/

#include "uisession.h"
#include "uiservercommon.h"

_LIT(KPanic, "CUiSession");
const TInt KBadRequest = 1;
const TInt KBadArguments = 2;
const TInt KAlreadyWaiting = 3;


// Factory method
CUiSession *CUiSession::NewL(RThread &aClientThread,
                             CUiServer &aServer,
                             MUiServerObserver &aObserver,
                             CTerminalControl &aTerminal) {
    CUiSession *self = new (ELeave) CUiSession(aClientThread, aServer,
                                               aObserver, aTerminal);
    CleanupStack::PushL(self);
    self->ConstructL();
    CleanupStack::Pop();
    return self;
}


// Constructor
CUiSession::CUiSession(RThread &aClientThread, CUiServer &aServer,
                       MUiServerObserver &aObserver,
                       CTerminalControl &aTerminal)
    : CSession(aClientThread),
      iServer(aServer),
      iClientThread(aClientThread),
      iObserver(aObserver),
      iTerminal(aTerminal) {
    iServer.AddSession();
}


// Second-phase constructor
void CUiSession::ConstructL() {
    iTerminal.SetObserver(this);
    iTerminal.ClearTerminal(KRgbBlack);
}


// Destructor
CUiSession::~CUiSession() {

    if ( iWaitingKey ) {
        iGetKeyMessage.Complete(KErrDied);
    }        
    
    iTerminal.RemoveObserver();
    iTerminal.ClearTerminal(KRgbBlack);
    iServer.RemoveSession();
}


// Service a request
void CUiSession::ServiceL(const RMessage &aMessage) {
    TRAPD(error, DispatchMessageL(aMessage));
    if ( error != KErrNone ) {
        aMessage.Complete(error);
    }
}


// Handle a message from a client
void CUiSession::DispatchMessageL(const RMessage &aMessage) {

    switch ( aMessage.Function() ) {
        case EUiReqDrawText:
            DrawTextL(aMessage);
            return;

        case EUiReqGetKey:
            GetKeyL(aMessage);
            return;

        case EUiErrorMessage:
            ErrorMessageL(aMessage);
            return;

        case EUiSetCursor:
            SetCursorL(aMessage);
            return;

        default:
            Panic(KPanic, KBadRequest);
    }    
}


// Draw text
void CUiSession::DrawTextL(const RMessage &aMessage) {

    // Read parameters
    TUiDrawTextBuf paramBuf;
    aMessage.ReadL(aMessage.Ptr0(), paramBuf);

    // Read text
    HBufC *textBuf = HBufC::NewLC(paramBuf().iTextLength);
    TPtr text = textBuf->Des();
    aMessage.ReadL(aMessage.Ptr1(), text);

    // Draw text
    iTerminal.DrawText(paramBuf().iX, paramBuf().iY, text, EFalse, EFalse,
                       paramBuf().iFgColor, paramBuf().iBgColor);

    CleanupStack::PopAndDestroy(); // textBuf
    aMessage.Complete(KErrNone);
}


// Get key
void CUiSession::GetKeyL(const RMessage &aMessage) {

    if ( iWaitingKey ) {
        Panic(KPanic, KAlreadyWaiting);
    }

    iWaitingKey = ETrue;
    iGetKeyMessage = aMessage;
}


// Error message
void CUiSession::ErrorMessageL(const RMessage &aMessage) {

    // Get the message
    const TAny *p = aMessage.Ptr0();
    TInt length = aMessage.Client().GetDesLength(p);
    HBufC *buf = HBufC::NewLC(length);
    TPtr ptr = buf->Des();
    aMessage.ReadL(p, ptr);

    // Send it to the observer
    iObserver.MusoErrorMessage(ptr);

    CleanupStack::PopAndDestroy(); // buf
    aMessage.Complete(KErrNone);
}


// Set cursor position
void CUiSession::SetCursorL(const RMessage &aMessage) {

    // Read parameters
    TUiCursorBuf buf;
    aMessage.ReadL(aMessage.Ptr0(), buf);

    iTerminal.SetCursor(buf().iX, buf().iY);

    aMessage.Complete(KErrNone);
}


// MTerminalObserver::TerminalSizeChanged()
void CUiSession::TerminalSizeChanged(TInt /*aWidth*/, TInt /*aHeight*/) {
}


// MTerminalObserver::KeyPressed()
void CUiSession::KeyPressed(TKeyCode aCode, TUint aModifiers) {

    // FIXME: Keystoke queue

    if ( iWaitingKey ) {
        TUiKeyBuf buf;
        buf().iKey = aCode;
        buf().iModifiers = aModifiers;
        iGetKeyMessage.WriteL(iGetKeyMessage.Ptr0(), buf);
        iGetKeyMessage.Complete(KErrNone);
        iWaitingKey = EFalse;
    }
}
